%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                      %%
%%  FlexClock                                                           %%
%%                                                                      %%
%%  (c) 2000 Universit catholique de Louvain.  All Rights Reserved.    %%
%%  The development of this software is supported by the PIRATES        %%
%%  project at the Universit catholique de Louvain.  This file is      %%
%%  subject to the general Mozart license.                              %%
%%                                                                      %%
%%  Author: Donatien Grolaux                                            %%
%%                                                                      %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

declare

[QTk]={Module.link ["x-oz://system/wp/QTk.ozf"]}


local

   %% NotifyClock sets procedures to call each second with the current time.
   
   NotifyClock
   local
      Clocks={NewCell nil}
      proc{Loop}
	 T={OS.localTime}
      in
	 for I in {Access Clocks} do {I T} end
	 {Delay 900} %% slightly less than 1000 to avoid missing a second.
	 {Loop}
      end
   in
      proc{NotifyClock P}
	 {Assign Clocks P|{Access Clocks}}
      end
      thread
	 {Loop}
      end
   end

   %% Formatting functions for displaying the time and date in several
   %% textual formats
   
   fun{TwoPos I}
      if I<10 then "0"#I else I end
   end

   fun{FmtTime T}
      {TwoPos T.hour}#":"#{TwoPos T.min}
   end

   fun{FmtTimeS T}
      {FmtTime T}#":"#{TwoPos T.sec}
   end

   fun{FmtDate T}
      {TwoPos T.mDay}#"/"#{TwoPos T.mon+1}
   end

   fun{FmtDateY T}
      {FmtDate T}#"/"#(1900+T.year)
   end

   fun{FmtDay T}
      {List.nth ["Monday" "Tuesday" "Wednesday" "Thursday" "Friday" "Saturday" "Sunday"] T.wDay}
   end

   fun{FmtMonth T}
      {List.nth ["January" "February" "March" "April" "May" "June" "July" "August"
		 "September" "October" "November" "December"] T.mon+1}
   end

   H1 H2 H3 H4 H5 H6 

   %% Definition of views. A view is defined by
   %% a) a record describing its GUI (in this case all views are composed of a single label)
   %% b) a procedure to update the displayed information according to the current time
   %% c) the minimal size required to correctly display the view
   %% A view definition is embedded in a single record, and all views are declared in the following list :
   
   ViewList=[r(refresh:proc{$ T} {H1 set(text:{FmtTime T})} end
	       desc:label(handle:H1 glue:nswe bg:white)
	       surface:40#10)
	     r(refresh:proc{$ T} {H2 set(text:{FmtTimeS T})} end
	       desc:label(handle:H2 glue:nswe bg:white)
	       surface:80#10)
	     r(refresh:proc{$ T} {H3 set(text:{FmtTime T}#'\n'#{FmtDate T})} end
	       desc:label(handle:H3 glue:nswe bg:white)
	       surface:40#30)
	     r(refresh:proc{$ T} {H4 set(text:{FmtTimeS T}#'\n'#{FmtDateY T})} end
	       desc:label(handle:H4 glue:nswe bg:white)
	       surface:80#30)
	     r(refresh:proc{$ T} {H5 set(text:{FmtTimeS T}#'\n'#{FmtDay T}#", "#{FmtDateY T})} end
	       desc:label(handle:H5 glue:nswe bg:white)
	       surface:130#30)
	     r(refresh:proc{$ T} {H6 set(text:{FmtTimeS T}#'\n'#{FmtDay T}#", "#
					 T.mDay#" "#{FmtMonth T}#" "#(1900+T.year))} end
	       desc:label(handle:H6 glue:nswe bg:white)
	       surface:180#30)]

   %% Construction of the window. The window is filled with a placeholder. A placeholder widget
   %% is a container widget that can change the displayed widget at any time as long as the window exist.
   
   P
   Window={QTk.build td(title:"FlexClock demo" placeholder(handle:P width:1 height:1 glue:nswe))}
   
   %% First, all views are placed once in the placeholder. When it is done, any view can be displayed
   %% again in the placeholder by a single command using the view's handle (see Place below).
   
   Views={List.map ViewList
	  fun{$ R}
	     Width#Height=R.surface
	  in
	     {P set(R.desc)}
	     {NotifyClock R.refresh}
	     Width#Height#(R.desc).handle
	  end}

   %% This procedure calculates the best view to display according to the current window size.
   %% The list of all views is parsed, the chosen view is the one that is
   %% 1) viewable (widthOfView < withOfWindow and heightOfView < heightOfWindow)
   %% 2) minimizes the distance between the bottom right corner of the view and the
   %% bottom right corner of the window (i.e. minimizes the norm 2 between these two points).
   
   proc{Place}
      WW={QTk.wInfo width(P)}
      WH={QTk.wInfo height(P)}
      _#Handle={List.foldRInd Views
		fun{$ I W#H#Handle Max#CH}
		   This=(W-WW)*(W-WW)+(H-WH)*(H-WH)
		in
		   if W<WW andthen H<WH andthen
		      (Max==inf orelse This<Max) then
		      This#Handle
		   else
		      Max#CH
		   end
		end
		inf#local (_#_#H)|_=Views in H end}
   in
      {P set(Handle)}
   end

   %% Each time the window is resized, the procedure above must be called to set the correct view
   %% according to the new size of the window. This is done by binding the <configure> event of QTk to
   %% execute the above procedure.
   
   {P bind(event:'<Configure>'
	   action:Place)}
   
in
   %% Window is now set up, but still hidden.
   {Window show}
   {Window wait}
end
   
